import storch
from torch.distributions import Normal, Categorical
import torch

torch.manual_seed(0)

mu_prior = torch.tensor([2.0, -3.0], requires_grad=True)
theta = torch.tensor([4.0, 5])

lax_method = storch.LAX("mu", in_dim=2)
expect = storch.Expect("k")
score_method = storch.ScoreFunction("white_noise_1", n_samples=2)
infer_method = storch.Infer("white_noise_2", Normal)


def loss(v):
    return torch.nn.MSELoss(reduction="none")(v, theta).mean(dim=-1)


mu = lax_method(Normal(mu_prior, 1))
k = expect(
    Categorical(
        probs=torch.tensor([[0.1, 0.3, 0.6], [0.1, 0.8, 0.1]], requires_grad=True)
    ),
)

agg_v = 0.0
s1 = 1.0
for i in range(2):
    k1, k2 = 0, 0
    if i == 1:
        k1 = k[:, 0]
        k2 = k[:, 1]
    s1 = score_method(Normal(mu + k1, 1))
    s2 = infer_method(Normal(-mu + s1 * k2, 1))
    # plus = lambda a, b: a + b
    # plus = storch.deterministic(plus)
    agg_v = agg_v + s1 + s2 * mu
    storch.add_cost(loss(agg_v), "loss")

storch.backward(debug=False, print_costs=True)



import storch
from torch.distributions import Normal
import torch
from storch.util import has_backwards_path

mu = torch.tensor([1., 0.2, 84.3], requires_grad=True)
theta = torch.tensor([-5., 4.2, 4.3])


@storch.stochastic
def white_noise(mu):
    return storch.sample(Normal(mu, 1)), torch.tensor([4., 3.])
print(white_noise(mu)[0])

# @storch.stochastic
# def white_noise(mu):
#     # Should crash: calls a stochastic context within a stochastic context
#     return white_noise(storch.sample(Normal(mu, 1)))
# s, c = white_noise(mu)


a = torch.tensor(0.4, requires_grad=True)
b = torch.tensor(0.5, requires_grad=True)
c = torch.tensor(0.6, requires_grad=True)

# d = a + b
# e = d * c
# print(has_differentiable_path(d, a)) # True
# print(has_differentiable_path(d, c)) # False
# print(has_differentiable_path(e, d)) # True
# print(has_differentiable_path(d, e)) # False
# print(has_differentiable_path(e, c)) # True
# print(has_differentiable_path(e, b)) # True


from typing import Optional

import storch
import torch
from torch.distributions import Distribution, Gumbel
import itertools


class SampleWithoutReplacementMethod(storch.Method):
    def __init__(self, plate_name: str, k: int):
        super().__init__(plate_name)
        self.k = k
        self.log_probs = None
        self.perturbed_log_probs = None

    def reset(self):
        self.log_probs = None
        self.perturbed_log_probs = None

    def _sample_tensor(
        self, distr: Distribution, parents: [storch.Tensor], plates: [storch.Plate]
    ) -> (torch.Tensor, int):
        # Perform stochastic beam search given the log-probs and perturbed log probs so far.
        # Samples k values from this distribution so that all total configurations are unique.
        # TODO: We have to keep track what samples are taken at each step for the backward pass.
        # Why? We need to think about what samples are discarded at some point because they are pruned away.
        # In the estimator they will still appear! So we'll have to think about that. They don't deserve a gradient
        # as they are only partial configurations and thus we don't know their loss.
        samples, self.log_probs, self.perturbed_log_probs = stochastic_beam_search(
            distr,
            self.k,
            len(plates),
            self.plate_name,
            self.log_probs,
            self.perturbed_log_probs,
        )
        return samples


def log1mexp(a: torch.Tensor) -> torch.Tensor:
    """See appendix A of http://jmlr.org/papers/v21/19-985.html.
    Numerically stable implementation of log(1-exp(a))"""
    r = torch.zeros_like(a)
    c = -0.693
    r[a > c] = (-a[a > c].expm1()).log()
    r[a <= c] = (-a[a <= c].exp()).log1p()
    return r


def log1pexp(a: torch.Tensor) -> torch.Tensor:
    """See appendix A of http://jmlr.org/papers/v21/19-985.html
    Numerically stable implementation of log(1+exp(a))
    TODO: The paper says there is a different case for a > 18... but the definition is invalid. need to check."""

    return (a.exp()).log1p()
    # r = torch.zeros_like(a)
    # c = 18
    # r[a >= c] = (-a[a>= c].exp()).log1p()
    # return r


def stochastic_beam_search(
    distribution: Distribution,
    k: int,
    amt_plates: int,
    # TODO: What is the plate dim? Should be the plate dim in the support tensor.
    plate_dim: str,
    log_probs: Optional[torch.Tensor],
    perturbed_log_probs: Optional[torch.Tensor],
) -> (torch.Tensor, torch.Tensor, torch.Tensor):
    """
    Implements Ancestral Gumbel-Top-k sampling with m=k, known as Stochastic Beam Search: http://jmlr.org/papers/v21/19-985.html.
    Sample from the distribution k items sequentially over the independent dimensions without replacement.
    Phi: Pass from method. Keeps track of current value of the perturbed log-probabilities. At storch.reset(), re-initialize.
    :param distribution:
    :param k:
    :param plate_dim Set to None if no dimension yet
    :return:
    """

    # TODO: This code is mostly taken from Expect(). May need to refactor to join these operations
    if not distribution.has_enumerate_support:
        raise ValueError(
            "Can only perform stochastic beam search for distributions with enumerable support."
        )

    support: torch.Tensor = distribution.enumerate_support(expand=True)

    sizes = support.shape[
        amt_plates + 1 : len(support.shape) - len(distribution.event_shape)
    ]

    ranges = []
    for size in sizes:
        ranges.append(range(size))

    support_non_expanded: torch.Tensor = distribution.enumerate_support(expand=False)
    yv_log_probs = distribution.log_prob(support_non_expanded)

    # Sample independent tensors in sequence
    # TODO: Check the shapes thoroughly here. There is the k dimension, which is the k best from the previous bfs layer
    # Then there is the current bfs layer, which is of shape |D_yv|. 
    samples = torch.zeros_like(support)
    if not plate_dim:
        samples.unsqueeze(0)
        plate_dim = 0
    for indices in itertools.product(ranges):
        if not log_probs:
            log_probs = yv_log_probs
            # First condition on max being 0:
            perturbed_log_probs = 0.0
        else:
            # Dunno how to shape this right, but this should be k times |support|=|D_yv|
            log_probs = log_probs + yv_log_probs
        # Sample |D_yv| Gumbel variables
        gumbel_d = Gumbel(loc=log_probs, scale=1.0)
        G_yv = gumbel_d.rsample()

        # Condition the Gumbel samples on the maximum of previous samples
        # TODO: again the shaping. Should take the maximum over |D_yv|
        Z = G_yv.max(-1)
        T = perturbed_log_probs
        vi = T - G_yv + log1mexp(G_yv - Z)
        cond_G_yv = T - vi.relu() - log1pexp(-vi.abs())

        # Select the k best
        # TODO: Argtop should be over both the k and the |D_yv| dimension, but is currently only over the last
        perturbed_log_probs, arg_top = torch.topk(cond_G_yv, k, dim=-1)
        log_probs = log_probs[arg_top]
        # Definitely wrong this shaping here, but that's the idea
        samples[..., *indices] = support_non_expanded[arg_top]
    return samples, log_probs, perturbed_log_probs



import torch
a = torch.tensor([[[[1., .4,.6, .9], [5., 4., 3., .2]]]])
b = torch.tensor([[[[True, False, True, False], [True, False, True, False]]]])

import storch
print(torch.Tensor.__getitem__(a, b))
print(a[b])
a[b] = 10.
print(a)
from torch.distributions import Normal, Categorical, Poisson
from storch.exceptions import IllegalStorchExposeError

method = storch.method.Infer(Poisson)

d = Normal(0, 1)

s = method.sample("oeps", d)
# print(s>1)
try:
    if s:
        print("Oeps")
except IllegalStorchExposeError:
    print("Good!!")

if s == s:
    print("ahh good")

d = Poisson(4)
s = method.sample(d)

print(s)

i = 0
for _ in range(s):
    i += 1


from typing import List, Union, Optional, Tuple

import storch
import torch

from storch import deterministic


def _convert_indices(
    tensor: storch.Tensor, dims=List[Union[str, int]]
) -> (List[int], List[str]):
    conv_indices = []
    red_batches = []
    for index in dims:
        if isinstance(index, int):
            if index >= tensor.plate_dims or index < 0 and index >= -tensor.event_dims:
                conv_indices.append(index)
            else:
                print(tensor.shape, index)
                raise IndexError(
                    "Can only pass indexes for event dimensions."
                    + str(tensor)
                    + ". Index: "
                    + str(index)
                )
        else:
            conv_indices.append(tensor.get_plate_dim_index(index))
            red_batches.append(index)
    return tuple(conv_indices), red_batches


def mean(tensor: storch.Tensor, dims=List[Union[str, int]]) -> storch.Tensor:
    indices, reduced_batches = _convert_indices(tensor, dims)
    return storch.reduce(torch.mean, plates=reduced_batches)(tensor, indices)


def sum(tensor: storch.Tensor, dims=List[Union[str, int]]) -> storch.Tensor:
    indices, reduced_batches = _convert_indices(tensor, dims)
    return storch.reduce(torch.sum, plates=reduced_batches)(tensor, indices)


def _handle_inputs(
    tensor: torch.Tensor,
    plates: Optional[List[storch.Plate]],
    plate_names: Optional[List[str]],
) -> (storch.Tensor, List[storch.Plate]):
    if plates and plate_names:
        raise ValueError("Provide only one of plates and plate_names.")
    if not isinstance(tensor, storch.Tensor):
        if not plates or plate_names:
            raise ValueError("Make sure to pass plates when passing a torch.Tensor.")
        index_tensor = 0
        for plate in plates:
            if plate.n > 1:
                if tensor.shape[index_tensor] != plate.n:
                    raise ValueError(
                        "Received a tensor that does not align with the given plates."
                    )
                index_tensor += 1
        tensor = storch.Tensor(tensor, [], plates)
    else:
        if plate_names:
            plates = []
            for plate in tensor.plates:
                if plate.name in plate_names:
                    plates.append(plate)
        elif not plates:
            plates = tensor.plates
    return tensor, plates


def reduce_plates(
    tensor: torch.Tensor,
    *,
    plates: Optional[List[storch.Plate]] = None,
    plate_names: Optional[List[str]] = None,
    detach_weights=True,
) -> storch.Tensor:
    """
    Reduce the tensor along the given plates. This takes into account how different samples are weighted, and should
    nearly always be used instead of reducing plate dimensions using the mean or the sum.
    :param tensor: Tensor to reduce
    :param plates: Plates to reduce. Cannot be used together with plate_names
    :param plate_names: Names of plates to reduce. Cannot be used togeter with plates
    :param detach_weights: Whether to detach the weighting of the samples from the graph
    :return: The reduced tensor
    """
    tensor, plates = _handle_inputs(tensor, plates, plate_names)
    for plate in plates:
        if plate.n > 1:
            tensor = plate.reduce(tensor, detach_weights=detach_weights)
    return tensor


def variance(
    tensor: torch.Tensor,
    variance_plate: Union[storch.Plate, str],
    *,
    plates: Optional[List[storch.Plate]] = None,
    plate_names: Optional[List[str]] = None,
    detach_weights=True,
) -> storch.Tensor:
    """
    Compute the variance of the tensor along the plate dimensions. This takes into account how different samples are weighted.
    :param tensor: Tensor to compute variance over
    :param plates: Plates to reduce. Cannot be used together with plate_names
    :param plate_names: Names of plates to reduce. Cannot be used togeter with plates
    :param detach_weights: Whether to detach the weighting of the samples from the graph
    :return: The reduced tensor
    """
    tensor, plates = _handle_inputs(tensor, plates, plate_names)
    found_plate = False
    for plate in plates:
        if plate == variance_plate:
            found_plate = True
            break
        if plate.name == variance_plate:
            variance_plate = plate
            found_plate = True
            break
    if not found_plate:
        raise ValueError(
            "Should pass a variance_plate that is included in the passed plates."
        )
    mean = variance_plate.reduce(tensor, detach_weights=detach_weights)
    variance = ((tensor - mean) ** 2).sum(tensor.event_dim_indices())
    return reduce_plates(variance, detach_weights=detach_weights)


def grad(
    outputs,
    inputs,
    grad_outputs=None,
    retain_graph: Optional[bool] = None,
    create_graph: bool = False,
    only_inputs: bool = True,
    allow_unused: bool = False,
) -> Tuple[storch.Tensor, ...]:
    """
    Helper method for computing torch.autograd.grad on storch tensors. Returns storch Tensors as well.
    """
    args, _, _, _ = storch.wrappers._handle_args(
        True, False, outputs, inputs, grad_outputs
    )
    _outputs, _inputs, _grad_outputs = tuple(args)
    grads = torch.autograd.grad(
        _outputs,
        _inputs,
        grad_outputs=_grad_outputs,
        retain_graph=retain_graph,
        create_graph=create_graph,
        only_inputs=only_inputs,
        allow_unused=allow_unused,
    )
    storch_grad = []
    for i, grad in enumerate(grads):
        input = inputs[i]
        if isinstance(input, storch.Tensor):
            storch_grad.append(
                storch.Tensor(
                    grad, outputs + [input], input.plates, input.name + "_grad"
                )
            )
        else:
            storch_grad.append(storch.Tensor(grad, outputs, [], "grad"))
    return tuple(storch_grad)



from __future__ import annotations

from typing import Union, Any, Tuple, List, Optional

import storch
import torch
from collections.abc import Iterable, Mapping
from functools import wraps
from storch.exceptions import IllegalStorchExposeError

_context_stochastic = False
_context_deterministic = 0
_stochastic_parents = []
_context_name = None
_plate_links = []
_ignore_wrap = False

# TODO: This is_iterable thing is a bit annoying: We really only want to unwrap them if they contain storch
#  Tensors, and then only for some types. Should rethink, maybe. Is unwrapping even necessary if the base torch methods
#  are all overriden? Maybe, see torch.cat?
def is_iterable(a: Any):
    return (
        isinstance(a, Iterable)
        and not isinstance(a, torch.Tensor)
        and not isinstance(a, str)
        and not isinstance(a, torch.Storage)
    )


def _collect_parents_and_plates(
    a: Any, parents: [storch.Tensor], plates: [storch.Plate]
) -> int:
    if isinstance(a, storch.Tensor):
        parents.append(a)
        for plate in a.plates:
            if plate not in plates:
                plates.append(plate)
        return a.event_dims
    elif is_iterable(a):
        max_event_dim = 0
        for _a in a:
            max_event_dim = max(
                max_event_dim, _collect_parents_and_plates(_a, parents, plates)
            )
        return max_event_dim
    elif isinstance(a, Mapping):
        max_event_dim = 0
        for _a in a.values():
            max_event_dim = max(
                max_event_dim, _collect_parents_and_plates(_a, parents, plates)
            )
        return max_event_dim
    return 0


def _unsqueeze_and_unwrap(
    a: Any, multi_dim_plates: [storch.Plate], align_tensors: bool, event_dims: int
):
    if isinstance(a, storch.Tensor):
        tensor = a._tensor
        if not align_tensors:
            return tensor
        # Automatically **RIGHT** broadcast. Ensure each tensor has an equal amount of event dims by inserting dimensions to the right
        # TODO: What do we think about this design?
        if a.event_dims < event_dims:
            tensor = tensor[(...,) + (None,) * (event_dims - a.event_dims)]

        # It can be possible that the ordering of the plates does not align with the ordering of the inputs.
        # This part corrects this.
        amt_recognized = 0
        links: [storch.Plate] = list(a.multi_dim_plates())
        for i, plate in enumerate(multi_dim_plates):
            if plate in a.plates:
                if plate != links[amt_recognized]:
                    # The plate is also in the tensor, but not in the ordering expected. So switch that ordering
                    j = links.index(plate)
                    tensor = tensor.transpose(j, amt_recognized)
                    links[amt_recognized], links[j] = links[j], links[amt_recognized]
                amt_recognized += 1

        for i, plate in enumerate(multi_dim_plates):
            if plate not in a.plates:
                tensor = tensor.unsqueeze(i)
        return tensor
    elif is_iterable(a):
        l = []
        for _a in a:
            l.append(
                _unsqueeze_and_unwrap(_a, multi_dim_plates, align_tensors, event_dims)
            )
        if isinstance(a, tuple):
            return tuple(l)
        return l
    elif isinstance(a, Mapping):
        d = {}
        for k, _a in a.items():
            d[k] = _unsqueeze_and_unwrap(
                _a, multi_dim_plates, align_tensors, event_dims
            )
        return d
    else:
        return a


def _handle_args(unwrap=True, align_tensors=True, *args, **kwargs):
    parents: [storch.Tensor] = []
    plates: [storch.Plate] = []
    max_event_dim = max(
        # Collect parent tensors and plates
        _collect_parents_and_plates(args, parents, plates),
        _collect_parents_and_plates(kwargs, parents, plates),
    )
    # Get the list of plates with size larger than 1 for the unsqueezing of tensors
    multi_dim_plates = []
    for plate in plates:
        if plate.n > 1:
            multi_dim_plates.append(plate)
    if unwrap:
        # Unsqueeze and align batched dimensions so that batching works easily.
        unsqueezed_args = []
        for t in args:
            unsqueezed_args.append(
                _unsqueeze_and_unwrap(t, multi_dim_plates, align_tensors, max_event_dim)
            )
        unsqueezed_kwargs = {}
        for k, v in kwargs.items():
            unsqueezed_kwargs[k] = _unsqueeze_and_unwrap(
                v, multi_dim_plates, align_tensors, max_event_dim
            )
        return unsqueezed_args, unsqueezed_kwargs, parents, plates
    return args, kwargs, parents, plates


def _process_deterministic(
    o: Any, parents: [storch.Tensor], plates: [storch.Plate], name: str
):
    if o is None:
        return
    if isinstance(o, storch.Tensor):
        if o.stochastic:
            raise RuntimeError(
                "Creation of stochastic storch Tensor within deterministic context"
            )
        # TODO: Does this require shape checking? Parent/Plate checking?
        return o
    if isinstance(o, torch.Tensor):  # Explicitly _not_ a storch.Tensor
        t = storch.Tensor(o, parents, plates, name=name)
        return t
    raise NotImplementedError(
        "Handling of other types of return values is currently not implemented: ", o
    )


def _deterministic(
    fn,
    *,
    unwrap: bool = True,
    align_tensors=True,
    reduce_plates: Optional[Union[str, List[str]]] = None
):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        if storch.wrappers._context_stochastic:
            # TODO check if we can re-add this
            raise NotImplementedError(
                "It is currently not allowed to open a deterministic context in a stochastic context"
            )
            # if storch.wrappers._context_deterministic > 0:
            #     if is_cost:
            #         raise RuntimeError("Cannot call storch.cost from within a deterministic context.")

            # TODO: This is currently uncommented and it will in fact unwrap. This was required because it was, eg,
            # possible to open a deterministic context, passing distributions with storch.Tensors as parameters,
            # then doing computations on these parameters. This is because these storch.Tensors will not be unwrapped
            # in the deterministic context as the unwrapping only considers lists.
            # # We are already in a deterministic context, no need to wrap or unwrap as only the outer dependencies matter
            # return fn(*args, **kwargs)

        new_args, new_kwargs, parents, plates = _handle_args(
            unwrap, align_tensors, *args, **kwargs
        )

        if not parents:
            return fn(*args, **kwargs)
        args = new_args
        kwargs = new_kwargs

        storch.wrappers._context_deterministic += 1

        try:
            outputs = fn(*args, **kwargs)
        finally:
            storch.wrappers._context_deterministic -= 1

        if storch.wrappers._ignore_wrap:
            return outputs
        nonlocal reduce_plates
        if reduce_plates:
            if isinstance(reduce_plates, str):
                reduce_plates = [reduce_plates]
            plates = [p for p in plates if p.name not in reduce_plates]
        if is_iterable(outputs):
            n_outputs = []
            for o in outputs:
                n_outputs.append(
                    _process_deterministic(
                        o, parents, plates, fn.__name__ + str(len(n_outputs))
                    )
                )
            outputs = n_outputs
        else:
            outputs = _process_deterministic(outputs, parents, plates, fn.__name__)
        return outputs

    return wrapper


def deterministic(fn=None, *, unwrap=True, align_tensors=True):
    """
    Wraps the input function around a deterministic storch wrapper.
    This wrapper unwraps :class:`~storch.Tensor` objects to :class:`~torch.Tensor` objects, aligning the tensors
    according to the plates, then runs `fn` on the unwrapped Tensors.
    :param fn: Function to wrap.
    :param unwrap: Set to False to prevent unwrapping :classs:`~storch.Tensor` objects.
    :return: The wrapped function `fn`.
    """
    if fn:
        return _deterministic(fn, unwrap=unwrap, align_tensors=align_tensors)
    return lambda _f: _deterministic(_f, unwrap=unwrap, align_tensors=align_tensors)


def reduce(fn, plates: Union[str, List[str]]):
    """
        Wraps the input function around a deterministic storch wrapper.
        This wrapper unwraps :class:`~storch.Tensor` objects to :class:`~torch.Tensor` objects, aligning the tensors
        according to the plates, then runs `fn` on the unwrapped Tensors. It will reduce the plates given by `plates`.
        :param fn: Function to wrap.
        :param unwrap: Set to False to prevent unwrapping :classs:`~storch.Tensor` objects.
        :return: The wrapped function `fn`.
        """
    if storch._debug:
        print("Reducing plates", plates)
    return _deterministic(fn, reduce_plates=plates)


def _self_deterministic(fn, self: storch.Tensor):
    fn = deterministic(fn)

    @wraps(fn)
    def wrapper(*args, **kwargs):
        # Inserts the self object at the beginning of the passed arguments. In essence, it "fakes" the self reference.
        args = list(args)
        args.insert(0, self)
        return fn(*args, **kwargs)

    return wrapper


def _process_stochastic(
    output: torch.Tensor, parents: [storch.Tensor], plates: [storch.Plate]
):
    if isinstance(output, storch.Tensor):
        if not output.stochastic:
            # TODO: Calls _add_parents so something is going wrong here
            # The Tensor was created by calling @deterministic within a stochastic context.
            # This means that we have to conservatively assume it is dependent on the parents
            output._add_parents(storch.wrappers._stochastic_parents)
        return output
    if isinstance(output, torch.Tensor):
        t = storch.Tensor(output, parents, plates)
        return t
    else:
        raise TypeError(
            "All outputs of functions wrapped in @storch.stochastic "
            "should be Tensors. At " + str(output)
        )


def stochastic(fn):
    """
    Applies `fn` to the `inputs`. `fn` should return one or multiple `storch.Tensor`s.
    `fn` should not call `storch.stochastic` or `storch.deterministic`. `inputs` can include `storch.Tensor`s.
    :param fn:
    :return:
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        if (
            storch.wrappers._context_stochastic
            or storch.wrappers._context_deterministic > 0
        ):
            raise RuntimeError(
                "Cannot call storch.stochastic from within a stochastic or deterministic context."
            )

        # Save the parents
        args, kwargs, parents, plates = _handle_args(*args, **kwargs)

        storch.wrappers._plate_links = plates
        storch.wrappers._stochastic_parents = parents
        storch.wrappers._context_stochastic = True
        storch.wrappers._context_name = fn.__name__

        try:
            outputs = fn(*args, **kwargs)
        finally:
            storch.wrappers._plate_links = []
            storch.wrappers._stochastic_parents = []
            storch.wrappers._context_stochastic = False
            storch.wrappers._context_name = None

        # Add parents to the outputs
        if is_iterable(outputs):
            processed_outputs = []
            for o in outputs:
                processed_outputs.append(_process_stochastic(o, parents, plates))
        else:
            processed_outputs = _process_stochastic(outputs, parents, plates)
        return processed_outputs

    return wrapper


def _exception_wrapper(fn):
    def wrapper(*args, **kwargs):
        for a in args:
            if isinstance(a, storch.Tensor):
                raise IllegalStorchExposeError(
                    "It is not allowed to call this method using storch.Tensor, likely "
                    "because it exposes its wrapped tensor to Python."
                )
        return fn(*args, **kwargs)

    return wrapper


def _unpack_wrapper(fn):
    def wrapper(*args, **kwargs):
        new_args = []
        for a in args:
            if isinstance(a, storch.Tensor):
                new_args.append(a._tensor)
            else:
                new_args.append(a)
        return fn(*tuple(new_args), **kwargs)

    return wrapper



from typing import List

from examples.vae import VAE, main
import torch.nn as nn
import torch
import storch
from torch.distributions import Normal, Distribution
from pyro.ops.indexing import Vindex


class NormalVAE(VAE):
    def initialize_param_layers(
        self, latents: int, prev_layer: int
    ) -> (nn.Module, nn.Module):
        # *2 because we compute both the mean and stdev
        fc3 = nn.Linear(prev_layer, latents * 2)
        fc4 = nn.Linear(latents, prev_layer)
        return fc3, fc4

    def initialize_method(self, args) -> storch.method.Method:
        if args.method == "reparameterization":
            return storch.Reparameterization("z", n_samples=args.samples)
        elif args.method == "lax":
            return storch.LAX("z", n_samples=args.samples, in_dim=args.latents)
        elif args.method == "score":
            return storch.ScoreFunction(
                "z", n_samples=args.samples, baseline_factory=args.baseline
            )

    def prior(self, posterior: Distribution) -> Distribution:
        return Normal(torch.zeros_like(posterior.loc), torch.ones_like(posterior.scale))

    def variational_posterior(self, params) -> Distribution:
        mean, std = params
        return Normal(mean, std)

    def logits_to_params(self, logits, latents):
        # mean, std
        return logits[..., :latents], torch.exp(0.5 * logits[..., latents:])

    def name(self):
        return "normal_vae"


if __name__ == "__main__":
    main(NormalVAE)



from typing import Callable

from .wrappers import (
    deterministic,
    stochastic,
    reduce,
    _exception_wrapper,
    _unpack_wrapper,
)
from .tensor import Tensor, CostTensor, StochasticTensor, Plate
from .method import *
from .inference import backward, add_cost, reset, denote_independent, gather_samples
from .util import print_graph
from .storch import *
import storch.typing
import storch.nn
from inspect import isclass
from .excluded_init import (
    _excluded_init,
    _exception_init,
    _unwrap_only_init,
    _excluded_function,
)

_debug = False

# Wrap all (common) normal PyTorch functions. This is required because they call
# C code. The returned tensors from the C code can only automatically be wrapped
# by wrapping the functions that call the C code.
# Ignore the first 35 as they are not related to tensors
let_through = False
for m, v in torch.__dict__.items():
    if m == "_adaptive_avg_pool2d":
        let_through = True
    if not let_through or not isinstance(v, Callable):
        continue
    if m in _exception_init:
        torch.__dict__[m] = _exception_wrapper(v)
    elif m in _unwrap_only_init:
        torch.__dict__[m] = _unpack_wrapper(v)
    elif (
        not isclass(v)
        and not str.startswith(m, "_cufft_")
        and not str.startswith(m, "rand")
        and m not in _excluded_init
    ):
        torch.__dict__[m] = deterministic(v)
    else:
        continue
    # torch.__dict__[m].__module__ = "torch"

let_through = False
for m, v in torch.nn.functional.__dict__.items():
    if m == "conv1d":
        let_through = True
    if let_through and isinstance(v, Callable) and m not in _excluded_function:
        torch.nn.functional.__dict__[m] = deterministic(v)



from __future__ import print_function
import argparse
from typing import Type

import torch
import torch.utils.data
from torch import optim
from torchvision.utils import save_image
from examples.dataloader.data_loader import data_loaders
from storch import backward, Expect
import storch
from tensorboardX import SummaryWriter
from examples.vae.vae import VAE

# Reconstruction + KL divergence losses summed over all elements and batch
def loss_function(recon_x, x):
    bce = storch.nn.b_binary_cross_entropy(recon_x, x, reduction="sum")
    return bce


def train(epoch, model, train_loader, device, optimizer, args, writer):
    model.train()
    train_loss = 0
    for batch_idx, (data, _) in enumerate(train_loader):
        data = data.to(device)
        optimizer.zero_grad()
        storch.reset()

        # Denote the minibatch dimension as being independent
        data = storch.denote_independent(data.view(-1, 784), 0, "data")
        recon_batch, KLD, z = model(data)
        storch.add_cost(loss_function(recon_batch, data), "reconstruction")
        cond_log = batch_idx % args.log_interval == 0
        cost = backward()
        train_loss += cost.item()

        optimizer.step()
        if cond_log:
            step = 100.0 * batch_idx / len(train_loader)
            global_step = 100 * (epoch - 1) + step

            # Variance of expect method is 0 by definition.
            variances = {}
            if args.method != "expect":
                _consider_param = "probs"
                if args.latents < 3:
                    old_method = model.sampling_method
                    model.sampling_method = Expect("z")
                    optimizer.zero_grad()
                    recon_batch, _, z = model(data)
                    storch.add_cost(loss_function(recon_batch, data), "reconstruction")
                    backward()
                    expect_grad = z.grad[_consider_param]

                    optimizer.zero_grad()
                    model.sampling_method = old_method
                grads = {n: [] for n in z.grad}
                for i in range(args.variance_samples):
                    optimizer.zero_grad()
                    recon_batch, _, z = model(data)
                    storch.add_cost(loss_function(recon_batch, data), "reconstruction")
                    backward()
                    for param_name, grad in z.grad.items():
                        grads[param_name].append(grad)

                variances = {}
                for param_name, gradz in grads.items():
                    # Create a new independent dimension for the different gradient samples
                    grad_samples = storch.gather_samples(gradz, "variance")
                    # Compute the variance over this independent dimension
                    variances[param_name] = storch.variance(
                        grad_samples, "variance"
                    )._tensor
                    if param_name == _consider_param and args.latents < 3:
                        mean = storch.reduce_plates(
                            grad_samples, plate_names=["variance"]
                        )
                        mse = storch.reduce_plates(
                            (grad_samples - expect_grad) ** 2
                        ).sum()
                        bias = (storch.reduce_plates((mean - expect_grad) ** 2)).sum()
                        print("mse", mse._tensor.item())
                        # Should approach 0 when increasing variance_samples for unbiased estimators.
                        print("bias", bias._tensor.item())
                        writer.add_scalar("train/probs_bias", bias._tensor, global_step)
                        writer.add_scalar("train/probs_mse", mse._tensor, global_step)

            print(
                "Train Epoch: {} [{}/{} ({:.0f}%)]\tCost: {:.6f}\t Logits var {}".format(
                    epoch,
                    batch_idx * len(data),
                    len(train_loader.dataset),
                    step,
                    cost.item(),
                    variances,
                )
            )
            writer.add_scalar("train/ELBO", cost, global_step)
            for param_name, var in variances.items():
                writer.add_scalar("train/variance/" + param_name, var, global_step)
    avg_train_loss = train_loss / (batch_idx + 1)
    print("====> Epoch: {} Average loss: {:.4f}".format(epoch, avg_train_loss))
    return avg_train_loss


def test(epoch, model, test_loader, device):
    model.eval()
    test_loss = 0
    with torch.no_grad():
        for i, (data, _) in enumerate(test_loader):
            data = data.to(device)
            data = storch.denote_independent(data.view(-1, 784), 0, "data")
            recon_batch, KLD, _ = model(data)
            test_loss += (
                loss_function(recon_batch, data).detach_tensor()
            ).mean() + KLD.detach_tensor().mean()
            if i == 0:
                n = min(data.size(0), 8)
                # comparison = storch.cat([data[:n],
                #                       recon_batch.detach_tensor()[0].view(args.batch_size, 1, 28, 28)[:n]]) # Take the first sample (0)
                # deterministic(save_image)(comparison.cpu(),
                #          'results/reconstruction_' + str(epoch) + '.png', nrow=n)

    test_loss /= i + 1
    print("====> Test set loss: {:.4f}".format(test_loss))
    return test_loss


def main(vae: Type[VAE]):
    parser = argparse.ArgumentParser(description="VAE MNIST Example")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=128,
        metavar="N",
        help="input batch size for training (default: 128)",
    )
    parser.add_argument(
        "--epochs",
        type=int,
        default=10,
        metavar="N",
        help="number of epochs to train (default: 10)",
    )
    parser.add_argument(
        "--no-cuda", action="store_true", default=False, help="enables CUDA training"
    )
    parser.add_argument(
        "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
    )
    parser.add_argument(
        "--log-interval",
        type=int,
        default=10,
        metavar="N",
        help="how many batches to wait before logging training status",
    )
    parser.add_argument(
        "--method",
        type=str,
        default="gumbel",
        help="Method in {gumbel, gumbel_straight, score}",
    )
    parser.add_argument(
        "--baseline",
        type=str,
        default="none",
        help="What baseline to use for the score function.",
    )
    parser.add_argument(
        "--latents",
        type=int,
        default=20,
        help="How many latent variables with 10 categories to use",
    )
    parser.add_argument(
        "--samples", type=int, default=1, help="How large of a budget to use"
    )
    parser.add_argument(
        "--variance_samples",
        type=int,
        default=10,
        help="How many samples to use to compute the variance of the estimators.",
    )
    parser.add_argument("--dataset", type=str, default="fixedMNIST")
    parser.add_argument("--lr", type=float, default=1e-3)
    args = parser.parse_args()
    args.cuda = not args.no_cuda and torch.cuda.is_available()

    torch.manual_seed(args.seed)

    device = torch.device("cuda" if args.cuda else "cpu")

    train_loader, test_loader = data_loaders(args)

    model = vae(args).to(device)
    writer = SummaryWriter(comment=model.name() + "_" + args.method)
    print(args)
    writer.add_text("hyperparameters", str(args), global_step=0)

    optimizer = optim.Adam(model.parameters(), lr=args.lr)

    best_train_loss = 1000.0
    best_test_loss = 1000.0
    for epoch in range(1, args.epochs + 1):
        train_loss = train(epoch, model, train_loader, device, optimizer, args, writer)
        if train_loss < best_train_loss:
            best_train_loss = train_loss
        test_loss = test(epoch, model, test_loader, device)
        writer.add_scalar("test_loss", test_loss, 100 * epoch)
        writer.flush()
        if test_loss < best_test_loss:
            best_test_loss = test_loss
        # TODO: Sampling currently doesn't work because of model.prior requiring the posterior.
        # with torch.no_grad():
        #     im_sample = model.prior([args.latents]).sample((64,))
        #     im_sample = model.decode(im_sample).cpu()
        #     save_image(
        #         im_sample.view(64, 1, 28, 28), "results/sample_" + str(epoch) + ".png",
        #     )
    measures = {
        "hparams/best_train_loss": best_train_loss,
        "hparams/best_test_loss": best_test_loss,
        # "hparams/train_loss": train_loss, "hparams/test_loss": test_loss,
        "train/loss": train_loss,
        "test_loss": test_loss,
    }
    writer.add_hparams(vars(args), measures, global_step=100 * epoch)
    writer.close()



from abc import ABC, abstractmethod
from torch.distributions import Distribution, OneHotCategorical, Bernoulli, Categorical

from storch.tensor import CostTensor, StochasticTensor, Plate
import torch
from typing import Optional, Type, Union, Dict, List
from storch.util import (
    has_differentiable_path,
    get_distr_parameters,
    rsample_gumbel,
)
from storch.typing import DiscreteDistribution, BaselineFactory
import storch
from storch.method.baseline import MovingAverageBaseline, BatchAverageBaseline
import itertools


class Method(ABC, torch.nn.Module):
    @staticmethod
    def _create_hook(sample: StochasticTensor, name: str, plates: List[Plate]):
        accum_grads = sample.param_grads
        del sample  # Remove from hook closure for GC reasons

        def hook(*args):
            # For some reason, this args unpacking is required for compatbility with registring on a .grad_fn...?
            # TODO: I'm sure there could be something wrong here
            grad = args[-1]
            if isinstance(grad, tuple):
                grad = grad[0]

            # print(grad)
            if name in accum_grads:
                accum_grads[name] = storch.Tensor(
                    accum_grads[name]._tensor + grad, [], plates, name + "_grad"
                )
            else:
                accum_grads[name] = storch.Tensor(grad, [], plates, name + "_grad")

        return hook

    def __init__(self, plate_name):
        super().__init__()
        self._estimation_pairs = []
        self.register_buffer("iterations", torch.tensor(0, dtype=torch.long))
        self.plate_name = plate_name

    def forward(self, distr: Distribution) -> storch.tensor._StochasticTensorBase:
        return self.sample(distr)

    def sample(self, distr: Distribution) -> storch.tensor._StochasticTensorBase:
        # Unwrap the distributions parameters
        params: Dict[str, storch.Tensor] = get_distr_parameters(
            distr, filter_requires_grad=False
        )
        parents: Dict[str, torch.Tensor] = {}
        # If we are in an @stochastic context, external plates might exist.
        plates: [Plate] = storch.wrappers._plate_links.copy()
        for name, p in params.items():
            if isinstance(p, storch.Tensor):
                parents[name] = p
                # The sample should have the batch links of the parameters in the distribution, if present.
                for plate in p.plates:
                    if plate not in plates:
                        plates.append(plate)

        # Will not rewrap in @deterministic, because sampling statements will insert an additional dimensions in the
        # first batch dimension, causing the rewrapping statement to error as it violates the plating constraints.
        storch.wrappers._ignore_wrap = True
        tensor = self._sample_tensor(distr, parents, plates)
        storch.wrappers._ignore_wrap = False

        plate_size = tensor.shape[0]
        if tensor.shape[0] == 1:
            tensor = tensor.squeeze(0)

        s_tensor = StochasticTensor(
            tensor,
            storch.wrappers._stochastic_parents + list(parents.values()),
            self,
            plates,
            distr,
            len(params) > 0,
            plate_size,
            self.plate_name,
        )

        for name, param in params.items():
            # TODO: Possibly could find the wrong gradients here if multiple distributions use the same parameter?
            # This maybe requires copying the tensor hm...
            if param.requires_grad:
                hook_plates = []
                if isinstance(distr, OneHotCategorical) or isinstance(
                    distr, Categorical
                ):
                    if param is not distr._param:
                        continue
                    # We only care about the input parameter. Ie, it returns both probs and logits, but only
                    # the one the user used to create the Distribution is of interest.
                    # These distributions first normalize their logits/probs. This causes incorrect gradient statistics.
                    # By taking a step back in the computation graph, we retrieve the correct parameter.
                    if isinstance(param, storch.Tensor):
                        hook_plates = param.plates
                        param = param._tensor
                    to_hook = param.grad_fn.next_functions[0][0]
                elif isinstance(param, storch.Tensor):
                    hook_plates = param.plates
                    to_hook = param._tensor
                else:
                    to_hook = param
                to_hook.register_hook(self._create_hook(s_tensor, name, hook_plates))

        # Possibly change something in the tensor now that it is wrapped and registered in the graph.
        # Used for example to rsample in LAX, and in post_sample detaching the tensor so that it won't record gradients
        # in the normal forward pass.
        edited_sample = self.post_sample(s_tensor)

        # Register this sampling method as being used in this iteration so that we can reset this method after the iteration
        if self not in storch.inference._sampling_methods:
            storch.inference._sampling_methods.append(self)

        if edited_sample is not None:
            new_s_tensor = storch.tensor._StochasticTensorBase(
                edited_sample._tensor,
                [s_tensor],
                s_tensor.plates,
                s_tensor.name,
                None,
                s_tensor.distribution,
                False,
                s_tensor.n,
            )
            new_s_tensor.param_grads = s_tensor.param_grads
            return new_s_tensor
        return s_tensor

    def _estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Optional[storch.Tensor]:
        self._estimation_pairs.append((tensor, cost_node))
        return self.estimator(tensor, cost_node)

    def _update_parameters(self):
        self.iterations += 1
        self.update_parameters(self._estimation_pairs)
        self._estimation_pairs = []

    @abstractmethod
    def _sample_tensor(
        self, distr: Distribution, parents: [storch.Tensor], plates: [Plate]
    ) -> torch.Tensor:
        pass

    def estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Optional[storch.Tensor]:
        # Estimators should optionally return a torch.Tensor that is going to be added to the total cost function.
        # In the case of for example reparameterization, None is returned to denote that no cost function is added.
        # When adding a loss function, adds_loss should return True
        return None

    def update_parameters(
        self, result_triples: [(StochasticTensor, CostTensor)]
    ) -> None:
        pass

    def plate_weighting(
        self, tensor: storch.StochasticTensor
    ) -> Optional[storch.Tensor]:
        # Weight by the size of the sample. Overload this if your gradient estimation uses some kind of weighting
        # of the different events, like importance sampling or computing the expectation.
        # If None is returned, it is assumed the samples are iid monte carlo samples.
        return None

    def adds_loss(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        """
        Returns true if the estimator will add an additional loss function for the derivative of the parameters
        of the stochastic tensor with respect to the cost node.
        """
        return False

    def post_sample(self, tensor: storch.StochasticTensor) -> Optional[storch.Tensor]:
        return None

    def reset(self):
        """
        This function gets called whenever storchastic is reset. This is after storch.backward() or storch.reset() is
        called. Can be used to reset this methods state to some initial state that has to happen every iteration. 
        """
        pass


class MonteCarloMethod(Method):
    def __init__(self, plate_name: str, n_samples: int = 1):
        super().__init__(plate_name)
        self.n_samples = n_samples


class Infer(MonteCarloMethod):
    """
    Method that automatically chooses reparameterization if it is available, otherwise the score function.
    """

    def __init__(
        self, plate_name: str, distribution_type: Type[Distribution], n_samples: int = 1
    ):
        super().__init__(plate_name, n_samples)
        if distribution_type.has_rsample:
            self._method = Reparameterization(plate_name, n_samples)
        elif issubclass(distribution_type, OneHotCategorical) or issubclass(
            distribution_type, Bernoulli
        ):
            self._method = GumbelSoftmax(plate_name, n_samples)
        else:
            self._method = ScoreFunction(plate_name, n_samples)
        self._score_method = ScoreFunction(plate_name, n_samples)

    def _sample_tensor(
        self, distr: Distribution, parents: [storch.Tensor], plates: [Plate]
    ) -> torch.Tensor:
        return self._method._sample_tensor(distr, parents, plates)

    def estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Optional[torch.Tensor]:
        return self._score_method.estimator(tensor, cost_node)

    def update_parameters(
        self, result_triples: [(StochasticTensor, CostTensor, torch.Tensor)]
    ):
        self._method.update_parameters(result_triples)
        self._score_method.update_parameters(result_triples)

    def adds_loss(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        if has_differentiable_path(cost_node, tensor):
            # There is a differentiable path, so we will just use reparameterization here.
            return False
        else:
            # No automatic baselines. Use the score function.
            return True


class Reparameterization(MonteCarloMethod):
    """
    Can only be used for reparameterizable distributions.
    Default option for reparameterizable distributions.
    """

    def _sample_tensor(
        self, distr: Distribution, parents: [storch.Tensor], plates: [Plate]
    ) -> torch.Tensor:
        if not distr.has_rsample:
            raise ValueError(
                "The input distribution has not implemented rsample. If you use a discrete "
                "distribution, make sure to use eg GumbelSoftmax."
            )
        return distr.rsample((self.n_samples,))

    def adds_loss(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        if has_differentiable_path(cost_node, tensor):
            # There is a differentiable path, so we will just use reparameterization here.
            return False
        raise ValueError(
            "There is no differentiable path between the cost tensor and the stochastic tensor. "
            "We cannot use reparameterization. Use a different gradient estimator, or make sure your"
            "code is differentiable."
        )


class GumbelSoftmax(Reparameterization):
    """
    Method that automatically chooses between Gumbel Softmax relaxation and the score function depending on the
    differentiability requirements of cost nodes. Can only be used for reparameterizable distributions.
    Default option for reparameterizable distributions.
    """

    def __init__(
        self,
        plate_name: str,
        n_samples: int = 1,
        straight_through=False,
        initial_temperature=1.0,
        min_temperature=1.0e-4,
        annealing_rate=1.0e-5,
    ):
        super().__init__(plate_name, n_samples)
        self.straight_through = straight_through
        self.register_buffer("temperature", torch.tensor(initial_temperature))
        self.register_buffer("annealing_rate", torch.tensor(annealing_rate))
        self.register_buffer("min_temperature", torch.tensor(min_temperature))

    def _sample_tensor(
        self, distr: DiscreteDistribution, parents: [storch.Tensor], plates: [Plate],
    ) -> torch.Tensor:
        return rsample_gumbel(
            distr, self.n_samples, self.temperature, self.straight_through
        )

    def update_parameters(
        self, result_triples: [(StochasticTensor, CostTensor, torch.Tensor)]
    ):
        if self.training:
            self.temperature = torch.max(
                self.min_temperature, torch.exp(-self.annealing_rate * self.iterations)
            )


class ScoreFunction(MonteCarloMethod):
    def __init__(
        self,
        plate_name: str,
        n_samples: int = 1,
        baseline_factory: Optional[Union[BaselineFactory, str]] = "moving_average",
        **kwargs,
    ):
        super().__init__(plate_name, n_samples)
        self.baseline_factory: Optional[BaselineFactory] = baseline_factory
        if isinstance(baseline_factory, str):
            if baseline_factory == "moving_average":
                # Baseline per cost possible? So this lookup/buffer thing is not necessary
                self.baseline_factory = lambda s, c: MovingAverageBaseline(**kwargs)
            elif baseline_factory == "batch_average":
                if n_samples == 1:
                    raise ValueError(
                        "Batch average can only be used for n_samples > 1. "
                    )
                self.baseline_factory = lambda s, c: BatchAverageBaseline()
            elif baseline_factory == "none" or baseline_factory == "None":
                self.baseline_factory = None
            else:
                raise ValueError("Invalid baseline name", baseline_factory)

    def _sample_tensor(
        self, distr: Distribution, parents: [storch.Tensor], plates: [Plate]
    ) -> torch.Tensor:
        return distr.sample((self.n_samples,))

    def estimator(self, tensor: StochasticTensor, cost: CostTensor) -> storch.Tensor:
        log_prob = tensor.distribution.log_prob(tensor)
        if len(log_prob.shape) > tensor.plate_dims:
            # Sum out over the event shape
            log_prob = log_prob.sum(
                dim=list(range(tensor.plate_dims, len(log_prob.shape)))
            )

        if self.baseline_factory:
            baseline_name = "_b_" + tensor.name + "_" + cost.name
            if not hasattr(self, baseline_name):
                setattr(self, baseline_name, self.baseline_factory(tensor, cost))
            baseline = getattr(self, baseline_name)
            cost = cost - baseline.compute_baseline(tensor, cost)
        # print(cost)
        return log_prob * cost.detach()

    def adds_loss(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        return True


class Expect(Method):
    def __init__(self, plate_name: str, budget=10000):
        super().__init__(plate_name)
        self.budget = budget

    def _sample_tensor(
        self, distr: Distribution, parents: [storch.Tensor], plates: [Plate]
    ) -> torch.Tensor:
        # TODO: Currently very inefficient as it isn't batched
        if not distr.has_enumerate_support:
            raise ValueError(
                "Can only calculate the expected value for distributions with enumerable support."
            )
        support: torch.Tensor = distr.enumerate_support(expand=True)
        support_non_expanded: torch.Tensor = distr.enumerate_support(expand=False)
        expect_size = support.shape[0]

        batch_len = len(plates)
        sizes = support.shape[
            batch_len + 1 : len(support.shape) - len(distr.event_shape)
        ]
        amt_samples_used = expect_size
        cross_products = 1 if not sizes else None
        for dim in sizes:
            amt_samples_used = amt_samples_used ** dim
            if not cross_products:
                cross_products = dim
            else:
                cross_products = cross_products ** dim

        if amt_samples_used > self.budget:
            raise ValueError(
                "Computing the expectation on this distribution would exceed the computation budget."
            )

        enumerate_tensor = support.new_zeros(
            [amt_samples_used] + list(support.shape[1:])
        )
        support_non_expanded = support_non_expanded.squeeze().unsqueeze(1)
        for i, t in enumerate(
            itertools.product(support_non_expanded, repeat=cross_products)
        ):
            enumerate_tensor[i] = torch.cat(t, dim=0)
        return enumerate_tensor.detach()

    def plate_weighting(
        self, tensor: storch.StochasticTensor
    ) -> Optional[storch.Tensor]:
        # Weight by the probability of each possible event
        log_probs = tensor.distribution.log_prob(tensor)
        if log_probs.plate_dims < len(log_probs.shape):
            log_probs = log_probs.sum(
                dim=list(range(tensor.plate_dims, len(log_probs.shape)))
            )
        return log_probs.exp()



from abc import ABC, abstractmethod
import torch
import storch
from storch.tensor import StochasticTensor, CostTensor


class Baseline(ABC, torch.nn.Module):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def compute_baseline(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> torch.Tensor:
        pass


class MovingAverageBaseline(Baseline):
    # TODO: Moving Average baseline does not take surrounding plates into account
    def __init__(self, exponential_decay=0.95):
        super().__init__()
        self.register_buffer("exponential_decay", torch.tensor(exponential_decay))
        self.register_buffer("moving_average", torch.tensor(0.0))

    def compute_baseline(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> torch.Tensor:
        # TODO: Use reduce instead of mean
        avg_cost = cost_node.detach_tensor().mean().detach()
        self.moving_average = (
            self.exponential_decay * self.moving_average
            + (1 - self.exponential_decay) * avg_cost
        )
        return self.moving_average


class BatchAverageBaseline(Baseline):
    # Uses the means of the other samples
    def compute_baseline(
        self, tensor: StochasticTensor, costs: CostTensor
    ) -> torch.Tensor:
        if tensor.n == 1:
            raise ValueError(
                "Can only use the batch average baseline if multiple samples are used."
            )
        costs = costs.detach()
        sum_costs = storch.sum(costs, tensor.name)
        # TODO: Should reduce correctly
        baseline = (sum_costs - costs) / (tensor.n - 1)
        return baseline


class BatchExclusionBaseline(Baseline):
    # Uses the means of the other samples, as long as they are different (this is biased, uses the sampled value in the second expectation)
    def compute_baseline(
        self, tensor: StochasticTensor, costs: CostTensor
    ) -> torch.Tensor:
        if tensor.n == 1:
            raise ValueError(
                "Can only use the batch average baseline if multiple samples are used."
            )
        costs = costs.detach()
        sum_costs = storch.sum(costs, tensor.name)
        baseline = (sum_costs - costs) / (tensor.n - 1)
        return baseline



'''
Edited from https://github.com/pytorch/examples/blob/master/vae/main.py
'''

from __future__ import print_function
import argparse
import torch
import torch.utils.data
from torch import nn, optim
from torch.nn import functional as F
from torchvision import datasets, transforms
from torchvision.utils import save_image
from storch import sample, deterministic, cost, backward
import storch
from torch.distributions import Normal

torch.manual_seed(0)

parser = argparse.ArgumentParser(description='VAE MNIST Example')
parser.add_argument('--batch-size', type=int, default=128, metavar='N',
                    help='input batch size for training (default: 128)')
parser.add_argument('--epochs', type=int, default=10, metavar='N',
                    help='number of epochs to train (default: 10)')
parser.add_argument('--no-cuda', action='store_true', default=False,
                    help='enables CUDA training')
parser.add_argument('--seed', type=int, default=1, metavar='S',
                    help='random seed (default: 1)')
parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                    help='how many batches to wait before logging training status')
args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()

torch.manual_seed(args.seed)

device = torch.device("cuda" if args.cuda else "cpu")

kwargs = {'num_workers': 1, 'pin_memory': True} if args.cuda else {}
train_loader = torch.utils.data.DataLoader(
    datasets.MNIST('../data', train=True, download=True,
                   transform=transforms.ToTensor()),
    batch_size=args.batch_size, shuffle=True, **kwargs)
test_loader = torch.utils.data.DataLoader(
    datasets.MNIST('../data', train=False, transform=transforms.ToTensor()),
    batch_size=args.batch_size, shuffle=True, **kwargs)


class VAE(nn.Module):
    def __init__(self):
        super(VAE, self).__init__()

        self.fc1 = nn.Linear(784, 400)
        self.fc21 = nn.Linear(400, 20)
        self.fc22 = nn.Linear(400, 20)
        self.fc3 = nn.Linear(20, 400)
        self.fc4 = nn.Linear(400, 784)

    def encode(self, x):
        h1 = F.relu(self.fc1(x))
        return self.fc21(h1), self.fc22(h1)

    @deterministic
    def decode(self, z):
        h3 = storch.relu(self.fc3(z))
        return self.fc4(h3).sigmoid()

    def KLD(self, mu, logvar):
        return -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())

    def forward(self, x):
        mu, logvar = self.encode(x.view(-1, 784))
        # Here!
        KLD = self.KLD(mu, logvar)
        std = torch.exp(0.5 * logvar)
        dist = Normal(mu, std)
        # z = sample(dist, method=storch.method.ScoreFunction(), n=100)
        z = dist.sample()
        log_prob = dist.log_prob(z)
        return self.decode(z), KLD, z, log_prob.sum()


model = VAE().to(device)
optimizer = optim.Adam(model.parameters(), lr=1e-5)


# Reconstruction + KL divergence losses summed over all elements and batch
def loss_function(recon_x, x):
    BCE = storch.nn.b_binary_cross_entropy(recon_x, x.view(-1, 784), reduction="sum")
    print(BCE)
    return BCE


def train(epoch):
    model.train()
    storch.reset()
    train_loss = 0
    for batch_idx, (data, _) in enumerate(train_loader):
        data = data.to(device)
        optimizer.zero_grad()
        recon_batch, KLD, z, log_prob = model(data)
        print(log_prob)
        BCE = loss_function(recon_batch, data)
        loss = BCE * log_prob + KLD
        loss.backward()
        train_loss += loss.item()
        optimizer.step()
        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader),
                loss.item() / len(data)))

    print('====> Epoch: {} Average loss: {:.4f}'.format(
          epoch, train_loss / len(train_loader.dataset)))


def test(epoch):
    model.eval()
    test_loss = 0
    with torch.no_grad():
        for i, (data, _) in enumerate(test_loader):
            data = data.to(device)
            recon_batch, KLD, _ = model(data)
            test_loss += (loss_function(recon_batch, data).detach_tensor() + KLD.detach_tensor()).mean()
            if i == 0:
                n = min(data.size(0), 8)
                comparison = storch.cat([data[:n],
                                      recon_batch.detach_tensor()[0].view(args.batch_size, 1, 28, 28)[:n]]) # Take the first sample (0)
                deterministic(save_image)(comparison.cpu(),
                         'results/reconstruction_' + str(epoch) + '.png', nrow=n)

    test_loss /= len(test_loader.dataset)
    print('====> Test set loss: {:.4f}'.format(test_loss))

if __name__ == "__main__":
    for epoch in range(1, args.epochs + 1):
        train(epoch)
        test(epoch)
        with torch.no_grad():
            im_sample = torch.randn(64, 20).to(device)
            im_sample = model.decode(im_sample).cpu()
            save_image(im_sample.view(64, 1, 28, 28),
                       'results/sample_' + str(epoch) + '.png')


from storch.method.method import (
    GumbelSoftmax,
    ScoreFunction,
    Infer,
    Method,
    Expect,
    Reparameterization,
    MonteCarloMethod,
)
from storch.method.relax import RELAX, REBAR, LAX
from storch.method.baseline import Baseline, MovingAverageBaseline



from storch import Tensor


class BlackboxTensor(Tensor):
    def __init__(self, tensor, parents, plates):
        super().__init__(tensor, parents, plates)
        self.tensor = tensor
        self.parents = parents
        self.batch_links = plates



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="storchastic",  # Replace with your own username
    version="0.0.1",
    author="Emile van Krieken",
    author_email="e.van.krieken@vu.nl",
    description="Deep Learning for PyTorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HEmile/storchastic",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from __future__ import annotations
import torch
import storch
from torch.distributions import Distribution
from collections import deque
from typing import Union, List, Tuple, Dict, Iterable, Any, Callable
import builtins
from itertools import product
from typing import Optional
from torch import Size
from storch.exceptions import IllegalStorchExposeError
from storch.excluded_init import (
    _exception_tensor,
    _unwrap_only_tensor,
    _excluded_tensor,
)

# from storch.typing import BatchTensor

_int = builtins.int
_size = Union[Size, List[_int], Tuple[_int, ...]]

_torch_dict = None


class Plate:
    def __init__(self, name: str, n: int, weight: Optional[storch.Tensor] = None):
        self.weight = weight
        if weight is None:
            self.weight = torch.tensor(1.0 / n)
        self.name = name
        self.n = n

    def __eq__(self, other):
        if not isinstance(other, Plate):
            return False
        if self.name != other.name:
            return False
        if self.n != other.n:
            # TODO: This should maybe return an error...?
            return False
        if isinstance(self.weight, storch.Tensor):
            if not isinstance(other.weight, storch.Tensor):
                return False
            if self.weight._tensor is other.weight._tensor:
                return True
            return self.weight._tensor.__eq__(other.weight._tensor).all()
        if isinstance(other.weight, storch.Tensor):
            return False
        return True  # Weights are equal if self.n == other.n

    def __str__(self):
        return self.name + ", " + str(self.n)

    def __repr__(self):
        return (
            "("
            + self.name.__repr__()
            + ", "
            + self.n.__repr__()
            + ", "
            + self.weight.__repr__()
            + ")"
        )

    def reduce(self, tensor: storch.Tensor, detach_weights=True):
        plate_weighting = self.weight
        if detach_weights:
            plate_weighting = self.weight.detach()
        # Case: The weight is a single number. First sum, then multiply with the weight (usually taking the mean)
        if plate_weighting.ndim == 0:
            return storch.sum(tensor, [self.name]) * plate_weighting

        # Case: There is a weight for each plate which is not dependent on the other batch dimensions
        elif plate_weighting.ndim == 1:
            index = tensor.get_plate_dim_index(self.name)
            plate_weighting = plate_weighting[
                (...,) + (None,) * (tensor.ndim - index - 1)
            ]
            weighted_tensor = tensor * plate_weighting
            return storch.sum(weighted_tensor, [self.name])

        # Case: The weight is a vector of numbers equal to batch dimension. Assumes it is a storch.Tensor
        else:
            weighted_tensor = tensor * plate_weighting
            return storch.sum(weighted_tensor, [self.name])


class Tensor(torch.Tensor):
    def __init__(
        self,
        tensor: torch.Tensor,
        parents: [Tensor],
        plates: [Plate],
        name: Optional[str] = None,
    ):
        if isinstance(tensor, Tensor):
            raise TypeError(
                "storch.Tensors should be constructed with torch.Tensors, not other storch.Tensors."
            )
        plate_names = set()
        batch_dims = 0
        # Check whether this tensor does not violate the constraints imposed by the given batch_links
        for plate in plates:
            if plate.name in plate_names:
                raise ValueError(
                    "Plates contain two instances of same plate "
                    + plate.name
                    + ". This can be caused by different samples with the same name using a different amount of samples n or different weighting of the samples. Make sure that these samples use the same number of samples."
                )
            plate_names.add(plate.name)
            # plate length is 1. Ignore this dimension, as singleton dimensions should not exist.
            if plate.n == 1:
                continue
            if len(tensor.shape) <= batch_dims:
                raise ValueError(
                    "Got an input tensor with a shape too small for its surrounding batch. Violated at dimension "
                    + str(batch_dims)
                    + " and plate shape dimension "
                    + str(len(plates))
                    + ". Instead, it was "
                    + str(len(tensor.shape))
                )
            elif not tensor.shape[batch_dims] == plate.n:
                raise ValueError(
                    "Storch Tensors should take into account their surrounding plates. Violated at dimension "
                    + str(batch_dims)
                    + " and plate "
                    + plate.name
                    + " with size "
                    + str(plate.n)
                    + ". "
                    "Instead, it was "
                    + str(tensor.shape[batch_dims])
                    + ". Batch links: "
                    + str(plates)
                    + " Tensor shape: "
                    + str(tensor.shape)
                )
            batch_dims += 1

        self._name = name
        self._tensor = tensor
        self._parents = []
        for p in parents:
            # TODO: Should I re-add this?
            # if p.is_cost:
            #     raise ValueError("Cost nodes cannot have children.")
            differentiable_link = has_backwards_path(self, p)
            self._parents.append((p, differentiable_link))
            p._children.append((self, differentiable_link))
        self._children = []
        self.plate_dims = batch_dims
        self.event_shape = tensor.shape[batch_dims:]
        self.event_dims = len(self.event_shape)
        self.plates = plates

    @staticmethod
    def __new__(cls, *args, **kwargs):
        # Pass the input tensor to register this tensor in C. This will initialize an empty (0s?) tensor in the backend.
        # TODO: Does that mean it will require double the memory?
        # return super(Tensor, cls).__new__(cls, device=tensor.device)
        return super(Tensor, cls).__new__(cls)

    def __hash__(self):
        return object.__hash__(self)

    @property
    def name(self):
        return self._name

    @property
    def is_sparse(self):
        return self._tensor.is_sparse

    def __str__(self):
        t = (
            (self.name + ": " if self.name else "") + "Stochastic"
            if self.stochastic
            else ("Cost" if self.is_cost else "Deterministic")
        )
        return t + " " + str(self._tensor) + " Batch links: " + str(self.plates)

    def __repr__(self):
        return object.__repr__(self)

    @storch.deterministic
    def __eq__(self, other):
        return self.__eq__(other)

    def _walk(
        self,
        expand_fn,
        depth_first=True,
        only_differentiable=False,
        repeat_visited=False,
        walk_fn=lambda x: x,
    ):
        visited = set()
        if depth_first:
            S = [self]
            while S:
                v = S.pop()
                if repeat_visited or v not in visited:
                    yield walk_fn(v)
                    visited.add(v)
                    for w, d in expand_fn(v):
                        if d or not only_differentiable:
                            S.append(w)
        else:
            queue = deque()
            visited.add(self)
            queue.append(self)
            while queue:
                v = queue.popleft()
                yield walk_fn(v)
                for w, d in expand_fn(v):
                    if (repeat_visited or w not in visited) and (
                        d or not only_differentiable
                    ):
                        visited.add(w)
                        queue.append(w)

    def walk_parents(
        self,
        depth_first=True,
        only_differentiable=False,
        repeat_visited=False,
        walk_fn=lambda x: x,
    ):
        return self._walk(
            lambda p: p._parents,
            depth_first,
            only_differentiable,
            repeat_visited,
            walk_fn,
        )

    def walk_children(
        self,
        depth_first=True,
        only_differentiable=False,
        repeat_visited=False,
        walk_fn=lambda x: x,
    ):
        return self._walk(
            lambda p: p._children,
            depth_first,
            only_differentiable,
            repeat_visited,
            walk_fn,
        )

    def detach_tensor(self) -> torch.Tensor:
        return self._tensor.detach()

    @property
    def stochastic(self) -> bool:
        return False

    @property
    def is_cost(self) -> bool:
        return False

    @property
    def requires_grad(self) -> bool:
        return self._tensor.requires_grad

    @property
    def plate_shape(self) -> torch.Size:
        return self._tensor.shape[: self.plate_dims]

    @property
    def shape(self) -> torch.Size:
        return self._tensor.size()

    def is_cuda(self):
        return self._tensor.is_cuda

    @property
    def dtype(self):
        return self._tensor.dtype

    @property
    def layout(self):
        return self._tensor.layout

    @property
    def device(self):
        return self._tensor.device

    @property
    def grad(self):
        return self._tensor.grad

    def dim(self):
        return self._tensor.dim()

    def ndimension(self):
        return self._tensor.ndimension()

    @property
    def ndim(self):
        return self._tensor.ndim

    def register_hook(self, hook: Callable) -> Any:
        return self._tensor.register_hook(hook)

    def event_dim_indices(self):
        return range(self.plate_dims, self._tensor.dim())

    def get_plate(self, plate_name: str) -> Plate:
        for plate in self.plates:
            if plate.name == plate_name:
                return plate
        raise IndexError("Tensor has no such plate: " + plate_name + ".")

    def get_plate_dim_index(self, plate_name: str) -> int:
        for i, plate in enumerate(self.multi_dim_plates()):
            if plate.name == plate_name:
                return i
        raise IndexError(
            "Tensor has no such plate: "
            + plate_name
            + ". Alternatively, the dimension of this batch is 1."
        )

    def iterate_plate_indices(self) -> Iterable[List[int]]:
        ranges = list(map(lambda a: list(range(a)), self.plate_shape))
        return product(*ranges)

    def multi_dim_plates(self) -> Iterable[Plate]:
        for plate in self.plates:
            if plate.n > 1:
                yield plate

    def backward(
        self,
        gradient: Optional[Tensor] = None,
        keep_graph: bool = False,
        create_graph: bool = False,
        retain_graph: bool = False,
    ) -> None:
        raise NotImplementedError(
            "Cannot call .backward on storch.Tensor. Instead, register cost nodes using "
            "storch.add_cost, then use storch.backward()."
        )

    # region OperatorOverloads

    def __len__(self):
        return self._tensor.__len__()

    # TODO: Is this safe?
    def __index__(self):
        return self._tensor.__index__()

    # TODO: shouldn't this have @deterministic?
    def eq(self, other):
        return self.eq(other)

    def __getstate__(self):
        raise NotImplementedError(
            "Pickle is currently not implemented for storch tensors."
        )

    def __setstate__(self, state):
        raise NotImplementedError(
            "Pickle is currently not implemented for storch tensors."
        )

    def __and__(self, other):
        if isinstance(other, bool):
            raise IllegalStorchExposeError(
                "Calling 'and' with a bool exposes the underlying tensor as a bool."
            )
        return storch.deterministic(self._tensor.__and__)(other)

    def __bool__(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to boolean. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a boolean."
        )

    def __float__(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to float. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a float."
        )

    def __int__(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to int. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as an int."
        )

    def __long__(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to long. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a long."
        )

    def __nonzero__(self) -> builtins.bool:
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to boolean. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a boolean."
        )

    def __array__(self):
        self.numpy()

    def __array_wrap__(self):
        self.numpy()

    def numpy(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to numpy arrays. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a np.array."
        )

    def __contains__(self, item):
        raise IllegalStorchExposeError(
            "It is not allowed to expose storch tensors via in statements."
        )

    def __deepcopy__(self, memodict={}):
        raise NotImplementedError(
            "There is currently no deep copying implementation for storch Tensors."
        )

    def __iter__(self):
        raise NotImplementedError("Cannot currently iterate over storch Tensors.")

    def detach_(self) -> Tensor:
        raise NotImplementedError("In place detach is not allowed on storch tensors.")

    # endregion


for m in dir(torch.Tensor):
    v = getattr(torch.Tensor, m)
    if (
        isinstance(v, Callable)
        and m not in Tensor.__dict__
        and m not in object.__dict__
    ):
        if m in _exception_tensor:
            setattr(torch.Tensor, m, storch._exception_wrapper(v))
        elif m in _unwrap_only_tensor:
            setattr(torch.Tensor, m, storch._unpack_wrapper(v))
        elif m not in _excluded_tensor:
            setattr(torch.Tensor, m, storch.deterministic(v))

# Yes. This code is extremely weird. For some reason, when monkey patching torch.Tensor.__getitem__ and
# torch.Tensor.__setitem__, bizarre indexing bugs will happen that wouldn't happen when not monkey patching them.
# Unsqueezing the masking tensor sometimes seems to help...
# To do this, I also had to reset the torch.Tensor method to its original state. This bug should be fixed sometimes,
# as this is extremely messy code.
original_get = torch.Tensor.__getitem__


def wrap_get(a, b):
    try:
        return storch.deterministic(original_get)(a, b)
    except IndexError:
        if storch._debug:
            print(
                "Got indexing error at __getitem__. Trying to resolve this using the unsqueeze hack."
            )

        @storch.deterministic
        def _weird_wrap(a, b):
            if isinstance(b, torch.Tensor):
                b = b.unsqueeze(0)
            return original_get(a, b)

        torch.Tensor.__getitem__ = original_get
        o = _weird_wrap(a, b)
        torch.Tensor.__getitem__ = wrap_get
        return o


torch.Tensor.__getitem__ = wrap_get

original_set = torch.Tensor.__setitem__


def wrap_set(a, b, v):
    try:
        return storch.deterministic(original_set)(a, b, v)
    except IndexError:
        if storch._debug:
            print(
                "Got indexing error at __setitem__. Trying to resolve this using the unsqueeze hack."
            )

        @storch.deterministic
        def _weird_wrap(a, b, v):
            if isinstance(b, torch.Tensor):
                b = b.unsqueeze(0)
            return original_set(a, b, v)

        torch.Tensor.__setitem__ = original_set
        o = _weird_wrap(a, b, v)
        torch.Tensor.__setitem__ = wrap_set
        return o


torch.Tensor.__setitem__ = wrap_set


class CostTensor(Tensor):
    def __init__(self, tensor: torch.Tensor, parents, plate_links: [Plate], name):
        super().__init__(tensor, parents, plate_links, name)

    @property
    def is_cost(self) -> bool:
        return True


class IndependentTensor(Tensor):
    """
    Used to denote independencies on a Tensor. This could for example be the minibatch dimension. The first dimension
    of the input tensor is taken to be independent and added as a batch dimension to the storch system.
    """

    def __init__(
        self,
        tensor: torch.Tensor,
        parents: [Tensor],
        plates: [Plate],
        tensor_name: str,
        plate_name: str,
        weight: Optional[storch.Tensor],
    ):
        n = tensor.shape[0]
        for plate in plates:
            if plate.name == plate_name:
                raise ValueError(
                    "Cannot create independent tensor with name "
                    + plate_name
                    + ". A parent sample has already used"
                    " this name. Use a different name for this independent dimension."
                )
        # TODO: Weighting
        plates.insert(0, Plate(plate_name, n, weight))
        super().__init__(tensor, parents, plates, tensor_name)
        self.n = n

    # TODO: Should IndependentTensors be stochastic? Sometimes, like if it is denoting a minibatch, making them
    #  stochastic seems like it is correct. Are there other cases?
    def stochastic(self) -> bool:
        return True


class _StochasticTensorBase(Tensor):
    def __init__(
        self,
        tensor: torch.Tensor,
        parents: [Tensor],
        plates: [Plate],
        name: str,
        sampling_method: Optional[storch.Method],
        distribution: Distribution,
        requires_grad: bool,
        n: int,
    ):
        self.distribution = distribution
        super().__init__(tensor, parents, plates, name)
        self._requires_grad = requires_grad
        self.n = n
        self.sampling_method = sampling_method
        self.param_grads = {}
        self._grad = None

    @property
    def stochastic(self):
        return True

    @property
    # TODO: Should not manually override it like this. The stochastic "requires_grad" should be a different method, so
    # that the meaning of requires_grad is consistent everywhere
    def requires_grad(self):
        return self._requires_grad

    @property
    def grad(self):
        return self.param_grads


class StochasticTensor(_StochasticTensorBase):
    # TODO: Copy original tensor to make sure it cannot change using inplace
    def __init__(
        self,
        tensor: torch.Tensor,
        parents: [Tensor],
        sampling_method: storch.Method,
        plates: [Plate],
        distribution: Distribution,
        requires_grad: bool,
        n: int,
        name: str,
    ):
        for plate in plates:
            if plate.name == name:
                raise ValueError(
                    "Cannot create stochastic tensor with name "
                    + name
                    + ". A parent sample has already used this name. Use a different name for this sample."
                )

        # Compute the batch weighting and add the resulting new plate. We have to temporarily assign a dummy
        # batch_links here in case the code inside sampling_method.batch_weighting refers to tensor.batch_links.
        plates.insert(0, Plate(name, n, None))
        super().__init__(
            tensor,
            parents,
            plates,
            name,
            sampling_method,
            distribution,
            requires_grad,
            n,
        )
        batch_weighting = sampling_method.plate_weighting(self)
        plate = Plate(name, self.n, batch_weighting)
        self.plates[0] = plate
        if isinstance(batch_weighting, storch.Tensor):
            batch_weighting.plates[0] = plate


from storch.util import has_backwards_path



from typing import Dict, Optional, List, Tuple, Union

from pyro.distributions import (
    RelaxedOneHotCategoricalStraightThrough,
    RelaxedBernoulliStraightThrough,
)

from storch.tensor import (
    Tensor,
    CostTensor,
    StochasticTensor,
    Plate,
    _StochasticTensorBase,
)
from torch.distributions import (
    Distribution,
    RelaxedOneHotCategorical,
    Categorical,
    OneHotCategorical,
    Bernoulli,
    RelaxedBernoulli,
)
import torch


def print_graph(costs: [CostTensor]):
    nodes = topological_sort(costs)
    counters = {"s": 1, "c": 1, "d": 1}
    names = {}

    def get_name(node, names):
        if node in names:
            return names[node]
        if node.name:
            if node.name not in counters:
                counters[node.name] = 1
                name = node.name + "[" + str(0) + "]"
            else:
                name = node.name + "[" + str(counters[node.name]) + "]"
                counters[node.name] += 1
        elif node.stochastic:
            name = "s" + "[" + str(counters["s"]) + "]"
            counters["s"] += 1
        elif node.is_cost:
            name = "c" + "[" + str(counters["c"]) + "]"
            counters["c"] += 1
        else:
            name = "d" + "[" + str(counters["d"]) + "]"
            counters["d"] += 1
        names[node] = name
        return name

    for node in nodes:
        name = get_name(node, names)
        print(name, node)
    for node in nodes:
        name = get_name(node, names)
        for p, differentiable in node._parents:
            edge = "-D->" if differentiable else "-X->"
            print(get_name(p, names) + edge + name)


def get_distr_parameters(
    d: Distribution, filter_requires_grad=True
) -> Dict[str, torch.Tensor]:
    params = {}
    for k in d.arg_constraints:
        try:
            p = getattr(d, k)
            if isinstance(p, torch.Tensor) and (
                not filter_requires_grad or p.requires_grad
            ):
                params[k] = p
        except AttributeError:
            from storch import _debug

            if _debug:
                print(
                    "Attribute",
                    k,
                    "was not added because we could not get the attribute from the object.",
                )
            pass
    return params


def _walk_backward_graph(grad, depth_first=True):
    if not grad:
        return
    if depth_first:
        for t, _ in grad.next_functions:
            if not t:
                continue
            yield t
            for o in _walk_backward_graph(t, depth_first):
                yield o
    else:
        for t, _ in grad.next_functions:
            yield t
        for t, _ in grad.next_functions:
            for o in _walk_backward_graph(t, depth_first):
                yield o


def walk_backward_graph(tensor: torch.Tensor, depth_first=True):
    if not tensor.grad_fn:
        raise ValueError("Can only walk backward over graphs with a gradient function.")
    return _walk_backward_graph(tensor.grad_fn, depth_first)


def has_backwards_path(output: Tensor, input: Tensor, depth_first=False):
    """
    Returns true if the gradient functions of the torch.Tensor underlying output is connected to the input tensor.
    This is only run once to compute the possibility of links between two storch.Tensor's. The result is saved into the
    parent links on storch.Tensor's.
    :param output:
    :param input:
    :param depth_first: Initialized to False as we are usually doing this only for small distances between tensors.
    :return:
    """

    if isinstance(output, StochasticTensor):
        outputs = get_distr_parameters(output.distribution)
    else:
        outputs = {None: output._tensor}
    input = input._tensor
    if not input.requires_grad:
        return False
    for o in outputs.values():
        if isinstance(o, Tensor):
            o = o._tensor
        if o is input:
            # This can happen if the input is a parameter of the output distribution
            return True
        if not o.grad_fn:
            continue
        for p in walk_backward_graph(o, depth_first):
            if hasattr(p, "variable") and p.variable is input:
                return True
            elif input.grad_fn and p is input.grad_fn:
                return True
    return False


def has_differentiable_path(output: Tensor, input: Tensor):
    for c in input.walk_children(only_differentiable=True):
        if c is output:
            return True
    return False


def topological_sort(costs: [CostTensor]) -> [Tensor]:
    """
    Implements reverse kahn's algorithm
    :param costs:
    :return:
    """
    for c in costs:
        if not c.is_cost or c._children:
            raise ValueError(
                "The inputs of the topological sort should only contain cost nodes."
            )
    l = []
    s = costs.copy()
    edges = {}
    while s:
        n = s.pop()
        l.append(n)
        for (p, _) in n._parents:
            if p in edges:
                children = edges[p]
            else:
                children = list(map(lambda ch: ch[0], p._children))
                edges[p] = children
            c = -1
            for i, _c in enumerate(children):
                if _c is n:
                    c = i
                    break
            del children[c]
            if not children:
                s.append(p)
    return l


def tensor_stats(tensor: torch.Tensor):
    return "shape {} mean {:.3f} std {:.3f} max {:.3f} min {:.3f}".format(
        tuple(tensor.shape),
        tensor.mean().item(),
        tensor.std().item(),
        tensor.max().item(),
        tensor.min().item(),
    )


def reduce_mean(tensor: torch.Tensor, keep_dims: [int]):
    if len(keep_dims) == tensor.ndim:
        return tensor
    sum_out_dims = list(range(tensor.ndim))
    for dim in keep_dims:
        sum_out_dims.remove(dim)
    return tensor.mean(sum_out_dims)


def rsample_gumbel(
    distr: Distribution,
    n: int,
    temperature: torch.Tensor,
    straight_through: bool = False,
) -> torch.Tensor:
    if isinstance(distr, (Categorical, OneHotCategorical)):
        if straight_through:
            gumbel_distr = RelaxedOneHotCategoricalStraightThrough(
                temperature, probs=distr.probs
            )
        else:
            gumbel_distr = RelaxedOneHotCategorical(temperature, probs=distr.probs)
    elif isinstance(distr, Bernoulli):
        if straight_through:
            gumbel_distr = RelaxedBernoulliStraightThrough(
                temperature, probs=distr.probs
            )
        else:
            gumbel_distr = RelaxedBernoulli(temperature, probs=distr.probs)
    else:
        raise ValueError("Using Gumbel Softmax with non-discrete distribution")
    return gumbel_distr.rsample((n,))


def split(
    tensor: Tensor,
    plate: Plate,
    *,
    amt_slices: Optional[int] = None,
    slices: Optional[List[slice]] = None,
    create_plates=True
) -> Tuple[Tensor, ...]:
    """
    Splits the plate dimension on the tensor into several tensors and returns those tensors. Note: It removes the
    tensors from the computation graph and therefore should only be used when creating estimators, when logging or debugging, or
    if you know what you're doing.
    """
    if not slices:
        slice_length = int(plate.n / amt_slices)
        slices = []
        for i in range(amt_slices):
            slices.append(slice(i * slice_length, (i + 1) * slice_length))

    new_plates = tensor.plates.copy()

    index = tensor.get_plate_dim_index(plate.name)
    plates_index = new_plates.index(plate)
    if not create_plates and tensor.plate_dims != index + 1:
        for _plate in reversed(tensor.plates):
            if _plate.n > 1:
                # Overwrite old plate
                new_plates.remove(_plate)
                new_plates[plates_index] = _plate
                break
    empty_indices = [None] * (index - 1)
    sliced_tensors = []
    for _slice in slices:
        indices = empty_indices + [_slice]
        new_tensor = tensor._tensor[indices]
        if create_plates:
            n = _slice.stop - _slice.start
            final_plates = new_plates.copy()
            final_plates[plates_index] = Plate(plate.name, n)
            if n == 1:
                new_tensor = new_tensor.squeeze(index)
            new_tensor = Tensor(new_tensor, [], final_plates, tensor.name)
        else:
            new_tensor = Tensor(
                new_tensor.transpose(index, tensor.plate_dims - 1),
                [],
                new_plates,
                tensor.name,
            )
        sliced_tensors.append(new_tensor)
    return tuple(sliced_tensors)



from typing import Tuple, List

from examples.vae import VAE, main
import torch.nn as nn
import torch
import storch
from torch.distributions import OneHotCategorical, Distribution


class DiscreteVAE(VAE):
    def initialize_param_layers(
        self, latents: int, prev_layer: int
    ) -> (nn.Module, nn.Module):
        fc3 = nn.Linear(prev_layer, latents * 10)
        fc4 = nn.Linear(latents * 10, prev_layer)
        return fc3, fc4

    def initialize_method(self, args) -> storch.method.Method:
        if args.method == "gumbel":
            return storch.GumbelSoftmax("z", n_samples=args.samples)
        elif args.method == "gumbel_straight":
            return storch.GumbelSoftmax(
                "z", n_samples=args.samples, straight_through=True
            )
        elif args.method == "score":
            return storch.ScoreFunction(
                "z", n_samples=args.samples, baseline_factory=args.baseline
            )
        elif args.method == "expect":
            return storch.Expect("z")
        elif args.method == "relax":
            return storch.RELAX("z", n_samples=args.samples, in_dim=(args.latents, 10))
        elif args.method == "rebar":
            return storch.REBAR("z", n_samples=args.samples)
        elif args.method == "relax_rebar":
            return storch.RELAX(
                "z", n_samples=args.samples, in_dim=(args.latents, 10), rebar=True
            )
        else:
            raise ValueError("Invalid method passed to program arguments.")

    def prior(self, posterior: Distribution):
        return OneHotCategorical(probs=torch.ones_like(posterior.probs) / 10.0)

    def variational_posterior(self, logits: torch.Tensor):
        return OneHotCategorical(probs=logits.softmax(dim=-1))
        # return OneHotCategorical(logits=logits)

    def logits_to_params(self, logits, latents):
        return logits.reshape(logits.shape[:-1] + (latents, 10))

    def shape_latent(self, z, latents):
        return z.reshape(z.shape[:-2] + (latents * 10,))

    def name(self):
        return "discrete_vae"


if __name__ == "__main__":
    main(DiscreteVAE)
